/*
Copyright (c) 2001 by Matthias S. Benkmann <m.s.b@gmx.net>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; version 2
of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//KNOWN BUG: ParseString can not parse \xNN
//This means that dependencies on files with embedded non-printable chars
//other than the usual \t,\n,... will not be detected.

const char* USAGE=
"parsetrace - program for extracting dependency info from system call traces\n"
"Copyright (c) 2001 by Matthias Benkmann\n"
"\n"
"USAGE: parsetrace [-v] tracefile\n"
"\n"
"If -v is specified then parsetrace will list all dependency files in addition\n"
"to the owners.\n"
"tracefile should be created using\n"
"strace -q -o tracefile -f -F -e trace=open,execve,read <program-to-be-traced>\n";

#include <iostream>
#include <string>
#include <cstring>
#include <fstream>
#include <cctype>
#include <map>
#include <set>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <pwd.h>

using namespace std;

void ParseInt(string::const_iterator& I,string::const_iterator End,int& target)
{
  if (I==End) throw 42;
  target=0;
  int sign=+1;
  if (*I=='-') {++I; sign=-1;};
  while(I!=End && isdigit(*I))
  {
    target=target*10+(*I)-'0';
    ++I;
  };
  if (I!=End && !isspace(*I) && *I!=',' && *I!=')') throw 666;
  target*=sign;
};

void ParseString(string::const_iterator& I,string::const_iterator End,string& target)
{
  if (I==End) throw 404;
  target.erase(target.begin(),target.end());

  if (*I!='"') throw 13;
  ++I;
  while(I!=End && *I!='"')
  {
    char Cha[2]; Cha[1]=0; char& Ch=Cha[0];
    Ch=*I;
    if (Ch=='\\')
    {
      ++I;
      Ch=*I;
      switch(Ch)
      {
        case 'r': Ch='\r'; break; //carriage return (CR)
        case 'n': Ch='\n'; break; //line feed (LF)
        case 'a': Ch='\a'; break; //alert (BEL)
        case 'b': Ch='\b'; break; //backspace (BS)
        case 'f': Ch='\f'; break; //form feed (FF)
        case 't': Ch='\t'; break; //horizontal tab (HT)
        case 'v': Ch='\v'; break; //vertical tab (VT)
        case '0': Ch='\0'; break; //null (NUL)
      };
    };
    target.append(Cha); //stupid GCC 2.95.x doesn't know push_back(char)
    ++I;
  };
  if (I==End) throw 0;
};

//only accepts absolute paths for FName
void GetOwner(const string FName,string& Owner)
{
  Owner.erase(Owner.begin(),Owner.end());
  
  if (FName[0]!='/') return; 
  
  struct stat FStat;
  int i=stat(FName.c_str(),&FStat);
  if (i<0) return;
  uid_t UID=FStat.st_uid;
  struct passwd* PWDEnt=getpwuid(UID);
  if (PWDEnt==NULL) return;
  Owner=PWDEnt->pw_name;
};

int main(int argc,char* argv[])
{
try{
  if (argc<2)
  {
    cerr<<USAGE;
    return 1;
  };
  
  bool Verbose=false;
  if (strcmp("-v",argv[1])==0)
  {
    --argc;
    ++argv;
    Verbose=true;
  };
  
  ifstream In(argv[1]);
  if (!In)
  {
    cerr<<"Error opening "<<argv[1]<<'\n';
    return 1;
  };
  
  map<int,string> FNameByDescriptor;
  map<string,set<string> > ReadDeps;
  map<string,set<string> > ExecDeps;
  
  while(1)
  {
    string Line;
    getline(In,Line);
    if (In.fail())
    {
      if (!In.eof()) {cerr<<"Read error!\n"; return 1;};
      break;
    };
    
    string::const_iterator I=Line.begin();
    while(!isspace(*I++));
    while(isspace(*I)) ++I;
    string::const_iterator J=I;
    while(!(*J=='(')) ++J;
    
    string Command(I,J);

    I=J+1; //I points to 1st char of 1st argument to command;
    
    if (Command=="open")
    {
      string FName;
      ParseString(I,Line.end(),FName);
      while(*I++!='=');
      while(isspace(*I)) ++I;
      int fd;
      ParseInt(I,Line.end(),fd);
      FNameByDescriptor[fd]=FName;
//      cout<<"fd "<<fd<<" = "<<FName<<'\n';
    } else
    if (Command=="read")
    {
      int fd;
      ParseInt(I,Line.end(),fd);
      string Owner;
      GetOwner(FNameByDescriptor[fd],Owner);
      if (!Owner.empty()) ReadDeps[Owner].insert(FNameByDescriptor[fd]);
//      cout<<"Reading from fd "<<fd<<" ("<<FNameByDescriptor[fd]<<")\n";
    } else
    if (Command=="execve")
    {
      string FName;
      ParseString(I,Line.end(),FName);
      string Owner;
      GetOwner(FName,Owner);
      if (!Owner.empty()) ExecDeps[Owner].insert(FName);
//      cout<<"Executing "<<FName<<'\n';
    };
  };
  
  map<string,set<string> >::const_iterator I;
  map<string,set<string> >::const_iterator IEnd;
  
  cout<<"*********** read dependencies ***********\n";
  I=ReadDeps.begin();
  IEnd=ReadDeps.end();

  while(I!=IEnd) 
  {
    cout<< (*I).first <<'\n';
    if (Verbose)
    {
      set<string>::const_iterator J=(*I).second.begin();
      set<string>::const_iterator JEnd=(*I).second.end();
      while (J!=JEnd)
      {
        cout<<"\t"<< *J <<'\n';
        ++J;
      };
    };  
    ++I;
  };

  cout<<"\n";

  cout<<"*********** execve dependencies ***********\n";
  I=ExecDeps.begin();
  IEnd=ExecDeps.end();
  
  while(I!=IEnd) 
  {
    cout<< (*I).first <<'\n';
    if (Verbose)
    {
      set<string>::const_iterator J=(*I).second.begin();
      set<string>::const_iterator JEnd=(*I).second.end();
      while (J!=JEnd)
      {
        cout<<"\t"<< *J <<'\n';
        ++J;
      };
    };  
    ++I;
  };

  return 0;
}catch(int i)
{
  cerr<<"Unhandled integer exception: "<<i<<'\n';
  return 1;
}catch(...)
{
  cerr<<"Unhandled exception: <unknown>\n";
  return 1;
};
};
